#include <stdio.h>
#include <stdlib.h>
#include <time.h>

// Filter-Koeffizienten.  2 Taps.
float a0 = 0.5, a1 = - 0.5;

int main (int argc, char *argv[])
{
  if (argc != 3) { fputs ("usage: filter length iterations\n", stderr);  exit (1); }
  int size = atoi (argv[1]);  // Lnge des zu filternden Arrays
  int iter = atoi (argv[2]);  // Anzahl der Wiederholungen (zur besseren Zeitmessung)

  float *m, *x;  // Quell-Array
  float *y = calloc (size, sizeof (float));  // Ziel-Array fr sequentielle Version
  float *y2;  // Ziel-Array fr SIMD-Version

  int it, n, i;
  vector float *xp, *yp;

  // Permutation fr Shuffle-Befehl:
  vector unsigned char shp = {12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27};

  vector float a0v = {a0, a0, a0, a0};  // konstanter a0-Vektor fr SIMD-Multiplikation
  vector float a1v = {a1, a1, a1, a1};  // konstanter a1-Vektor fr SIMD-Multiplikation

  // zur Zeitmessung:
  clock_t t1, t2, t3;
  clock_t pms = CLOCKS_PER_SEC / 1000;

  // Allozieren der Arrays mit 16-Byte-Alignment (4 floats)
  posix_memalign ((void **)&m, 4 * sizeof (float), (size+4) * sizeof (float));
  x = m + 4;  // x lsst negative Indizes bis -3 zu
  posix_memalign ((void **)&y2, 4 * sizeof (float), (size+4) * sizeof (float));

  // Initialisierung der Quell-Daten
  for (n = 0; n < size; n ++)  x[n] = n % 13;

  t1 = clock ();  // Zeitpunkt vor sequentiellem Algorithmus

  for (it = 0; it < iter; it ++)  // mehrere Iteration zwecks Zeitmessung
    // Sequentieller Algorithmus
    for (n = 0; n < size; n ++)
      y[n] = x[n] * a0 + x[n-1] * a1;

  t2 = clock ();  // Zeitpunkt nach sequ. und vor SIMD-Algorithmus

  for (it = 0; it < iter; it ++)
  {
    // SIMD-Algorithmus
    xp = (vector float *) x;   // Pointer-Zugriff auf Array ist schneller
    yp = (vector float *) y2;  // %

    for (n = 0; n < size/4; n ++)  // size/4 weil 4 floats gleichzeitig bearbeitet werden
    {
      /* Filterung.  Der Vektor (x[-4],x[-3],x[-2],x[-1]) wird mit
         einem Shuffle-Befehl (vec_perm) aus xp[-1] =
         (x[-7],x[-6],x[-5],x[-4]) und xp[0] =
         (x[-3],x[-2],x[-1],x[0]) berechnet. */
      *yp = *xp * a0v + vec_perm (xp[-1], *xp, shp) * a1v;
      yp ++;  xp ++;  // Pointer um 16 Byte = 4 Floats = 1 Vektor erhhen
    }
  }

  t3 = clock ();  // Zeitpunkt nach SIMD-Algorithmus

  // Unterschied zwischen den zwei Algorithmen berechnen, um Fehler zu erkennen
  double error = 0.0;
  for (i = 0; i < size; i ++)
  {
    double diff = y[i] - y2[i];
    error += diff * diff;
  }
  error /= (double) size;

  // Ausgabe von Ausfhrungszeiten, Speedup und Fehler
  printf ( " plain time: %d ms\n vector time: %d ms\n speedup: %f\n error: %f\n"
         , (int) ((t2 - t1) / pms)
         , (int) ((t3 - t2) / pms)
         , (double) (t2 - t1) / (double) (t3 - t2)
         , error
         );

  // Speicher freigeben
  free (m);
  free (y);
  free (y2);

  return 0;
}
